/* Lazyread 1.6  -by Ryan Kulla */
/* HomePage: http://www30.brinkster.com/gt3world/myworld.html */
/* E-Mail: ambiod@sbcglobal.net */
/* License: You can freely distribute this program giving credit
            to the original author: Ryan Kulla. */

#include <curses.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DEFWORD "lazyread"
#define DEFSPEED 1000
#define BUFMAX 255
#define LR_VERSION "1.6"
#define OFF 0
#define ON 1
#define UC 1
#define LC 2

struct my_windows {
    WINDOW *scrollwin;
    WINDOW *statwin;
} *pstat, *pscroll;

char *filename = NULL;
char *progname;
char *special_word = NULL;
FILE *fp = NULL;
unsigned int Y = 0; 
unsigned int POS_CHANGED = 0;
unsigned int ok = 0;
unsigned int wflg = 0;
unsigned int VIEW_NORMAL = 0;
unsigned int PIPED = 1;
unsigned int ALONE = 0;
unsigned int CASE_FLAG = 0;
unsigned int USE_CASE = 0;
unsigned int SCROLL_LETTERS = 0;
unsigned int NO_STATUSBAR = 0;
unsigned int AUTO_PAUSE = 0;
unsigned int BEEP_OK = 0;
unsigned int WAS_IN_SHOW_INFO = 0;
unsigned int HIGHLIGHT_FLAG = 0;
unsigned long int THE_FILE_SIZE = 0;

void scan_command_line(int, char **);
void scroll_it(unsigned int, int, char *);
void usage(void);
void quit_cleanly(void);
void get_stats(unsigned long int, unsigned long int);
void highlight_word(char *, unsigned int, unsigned int, unsigned int, unsigned long int, unsigned long int);
void user_input(unsigned int *, unsigned int, unsigned int *, unsigned long int, unsigned long int);
void do_options(unsigned int, int, char *);
unsigned int get_key(void);
int char_check(char *);
long line_nums(FILE *);
void change_case(char *, int);
void char_scroll(FILE *, unsigned int);
void show_info(unsigned int, unsigned int, unsigned long int, unsigned long int);
int file_size(FILE *);
void start_editor(unsigned long int);

int main(int argc, char **argv)
{
    initscr(); 
    echo();
    cbreak();

    progname = argv[0];
 
    scan_command_line(argc, argv);
  
    refresh();
    endwin();
    return 0;
}

void scan_command_line(int argc, char **argv)
{
char speed[20], position[3], *argopt = NULL;//
unsigned int scroll_speed = DEFSPEED, opt; 

    for (opt = 1; opt < argc; opt++) {
        if (!strncmp(argv[opt], "-s", 2))
            if ((argopt = strchr(argv[opt], '\0'))) {
                if (!char_check(argopt + 1)) 
                    usage(); 
                strncpy(speed, argopt + 1, 20);
                scroll_speed = (unsigned int)atoi(speed);
            }
        if (!strncmp(argv[opt], "-pos", 4))
            if ((argopt = strchr(argv[opt], '\0'))) {
                if (!char_check(argopt + 1))
                    usage(); 
                strncpy(position, argopt + 1, 3);
                POS_CHANGED = (unsigned int)atoi(position);
            }
        if (!strncmp(argv[opt], "-f", 2)) {
            PIPED = 0;
            if ((argopt = strchr(argv[opt], '\0'))) { 
                if (strlen(argopt + 1) > BUFMAX)
                    usage(); 
                if (!(filename = (char *)malloc(strlen(argopt+1)+1))) {
                    perror("malloc()");
                    exit(EXIT_FAILURE);
                }
                strcpy(filename, argopt + 1);
                if (!(fp = fopen(filename, "r"))) {
                    refresh();
                    endwin();
                    perror("fopen()");
                    exit(EXIT_FAILURE);
                }
            }
        }
        if (!strncmp(argv[opt], "-w", 2))
            if ((argopt = strchr(argv[opt], '\0'))) {
                if (strlen(argopt + 1) > BUFMAX)
                    usage();
                wflg = 1;
                if (!(special_word = (char *)malloc(strlen(argopt+1)+1))) {
                    perror("malloc()");
                    exit(EXIT_FAILURE);
                }
                strcpy(special_word, argopt + 1);
            }
        if (!strncmp(argv[opt], "-n", 2))
            VIEW_NORMAL = 1;
        if (!strncmp(argv[opt], "-case", 5))
            if ((argopt = strchr(argv[opt], '\0'))) {
                if (!strncmp(argopt + 1, "upper", 5)) {
                    USE_CASE = UC;
                    CASE_FLAG = 1;
                }
                if (!strncmp(argopt + 1, "lower", 5)) {
                    USE_CASE = LC;
                    CASE_FLAG = 1;
                }
            }
        if (!strncmp(argv[opt], "-color", 6)) 
            if (has_colors() == FALSE) {
                endwin();
                printf("Your terminal doesn't support colors.\n");
                exit(EXIT_FAILURE);
            } 
            start_color();
            if ((argopt = strchr(argv[opt], '\0'))) 
                if (!strncmp(argopt + 1, "red", 3)) 
                    assume_default_colors(COLOR_RED, COLOR_BLACK);
                if (!strncmp(argopt + 1, "bgblue", 6))
                    assume_default_colors(COLOR_WHITE, COLOR_BLUE);
                if (!strncmp(argopt + 1, "blue", 4))
                    assume_default_colors(COLOR_BLUE, COLOR_BLACK);
                if (!strncmp(argopt + 1, "green", 5))
                    assume_default_colors(COLOR_GREEN, COLOR_BLACK);
                if (!strncmp(argopt + 1, "magenta", 7))
                    assume_default_colors(COLOR_MAGENTA, COLOR_BLACK);
                if (!strncmp(argopt + 1, "yellow", 6))
                    assume_default_colors(COLOR_YELLOW, COLOR_BLACK);
        if (!strncmp(argv[opt], "-no-stat-bar", 12))
            NO_STATUSBAR = 1;
        if (!strncmp(argv[opt], "-v", 2)) {
            printf("lazyread %s. By Ryan Kulla\n", LR_VERSION);
            endwin();
            exit(EXIT_SUCCESS);
        }
        if (!strncmp(argv[opt], "-u", 2)) 
            usage();
        if (!strncmp(argv[opt], "-a", 2))
            AUTO_PAUSE = 1;
        if (!strncmp(argv[opt], "-l", 2))
            SCROLL_LETTERS = 1;
        if (!strncmp(argv[opt], "-b", 2))
            BEEP_OK = 1;
    }
        do_options(scroll_speed, argc, argv[1]);
}

void do_options(unsigned int scroll_speed, int argc, char *filename_no_dash_f)
{
    if (!wflg)
        special_word = DEFWORD; 

    if (scroll_speed)
         scroll_it(scroll_speed, argc, filename_no_dash_f);
    else
        usage();
}

void scroll_it(unsigned int scroll_speed, int argc, char *filename_no_dash_f)
{
char buf[BUFMAX], *s;
unsigned int origspeed = scroll_speed, toggle = ON, blank = 0;
FILE *which = NULL;
unsigned long int total_lines = 0, line = 0;

    if (!(pscroll = (struct my_windows *)malloc(sizeof(struct my_windows)))) {
        perror("malloc");
        exit(EXIT_FAILURE);
    }

    Y = LINES - 2;
    if (POS_CHANGED)
        Y = POS_CHANGED;
    if ((Y > LINES - 2) || (Y < 1)) 
        usage();

    pscroll->scrollwin = newwin(LINES, 0, 0, 0);
    buf[0] = '\0';

    noecho();
    curs_set(0);

    if (!PIPED)  { /* if they used -f */
        which = fp;
        total_lines = line_nums(which);
    } else {
        if (argc >= 2) { /* not piped but not -f'd either */ 
            filename = filename_no_dash_f;
	    if (!(which = fopen(filename, "r"))) 
                usage();
            total_lines = line_nums(which);
            ALONE = 1;
        }
        else  /* piped */
            which = stdin;
    }

    THE_FILE_SIZE = file_size(which);

    if (SCROLL_LETTERS) 
        char_scroll(which, scroll_speed);
    /* else: */
    while ((s = fgets(buf, sizeof(buf), which))) { /* scroll time */
        flushinp();
        line++;
          
        /* allows you to see single line output */
        if (line == 1) 
            refresh(); 

        /* Don't print 2+ blank lines */
        if (!VIEW_NORMAL) {
            if (buf[0] == '\n') 
                blank++;
            else
                blank = 0;
            if (blank == 2) { 
                blank = 1; continue; 
            }
        }

        if (!NO_STATUSBAR)
            if (toggle) 
               get_stats(total_lines, line);

        highlight_word(buf, scroll_speed, origspeed, toggle, total_lines, line);

        if (CASE_FLAG)
            change_case(buf, USE_CASE);

        mvwprintw(pscroll->scrollwin, Y, 0, "%s", buf);
        scrollok(pscroll->scrollwin, TRUE);
        napms(scroll_speed);
        scroll(pscroll->scrollwin);
        wrefresh(pscroll->scrollwin);
        user_input(&scroll_speed, origspeed, &toggle, total_lines, line);
    } 
    fclose(which);
}

void char_scroll(FILE *which, unsigned int scroll_speed)
{
unsigned int i, line = 0;
char buf[BUFMAX], *s;
unsigned int origspeed = scroll_speed, toggle = ON;
unsigned long int total_lines = line_nums(which);
  
    while ((s = fgets(buf, sizeof(buf), which))) {
        line++;
        if (line == 1) refresh(); 
        highlight_word(buf, scroll_speed, origspeed, toggle, total_lines, line);
        if (CASE_FLAG)
            change_case(buf, USE_CASE);
        if (HIGHLIGHT_FLAG) {
            HIGHLIGHT_FLAG = 0;
            wclear(pscroll->scrollwin);
            wrefresh(pscroll->scrollwin);
            continue;
        }
        for (i = 0; buf[i] != '\0'; i++) {
            if (!NO_STATUSBAR)
                if (toggle)
                    get_stats(total_lines, line);
            if (WAS_IN_SHOW_INFO) {
                get_stats(total_lines, line); /* see status bar */
                WAS_IN_SHOW_INFO = 0;
            }
            if (POS_CHANGED)
                mvwprintw(pscroll->scrollwin, Y, i, "%c", buf[i]);
            else
                mvwprintw(pscroll->scrollwin, Y / 2, i, "%c", buf[i]);
            napms(scroll_speed); 
            wrefresh(pscroll->scrollwin);
            user_input(&scroll_speed, origspeed, &toggle, total_lines, line);
        }
        wclear(pscroll->scrollwin);
        wrefresh(pscroll->scrollwin);
    } 
}

int file_size(FILE *fp)
{
unsigned long int len = 0;

    fseek(fp, 0, SEEK_END);
    len = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    return(len);
}

void change_case(char *buf, int choice)
{
char *p;

    for (p = buf; *p != '\0'; p++) {
        if (choice == LC)
            *p = tolower(*p);
        else
            *p = toupper(*p);
    }
} 

void highlight_word(char *buf, unsigned int scroll_speed, unsigned int origspeed, unsigned int toggle, unsigned long int total_lines, unsigned long int line)
{
unsigned int i = 0;

    /* highlight the entire line special_word is on */
    if (strstr(buf, special_word)) {
        HIGHLIGHT_FLAG = 1;//
        wattrset(pscroll->scrollwin, A_BOLD);
        if (BEEP_OK)
            beep(); 

        /* print again to see it correctly */
        if (SCROLL_LETTERS) { 
            for (i = 0; buf[i] != '\0'; i++) {
                if (!NO_STATUSBAR)
                    if (toggle)
                        get_stats(total_lines, line);
                if (POS_CHANGED)
                    mvwprintw(pscroll->scrollwin, Y, i, "%c", buf[i]);
                else
                    mvwprintw(pscroll->scrollwin, Y / 2, i, "%c", buf[i]);
                user_input(&scroll_speed, origspeed, &toggle, total_lines, line);
                napms(scroll_speed);
                wrefresh(pscroll->scrollwin);
            }
        }
        else
            mvwprintw(pscroll->scrollwin, Y, 0, "%s", buf);

        if (AUTO_PAUSE) {
            wrefresh(pscroll->scrollwin);
            nodelay(stdscr, FALSE);
            getch();
        }
    }
    else
        wattrset(pscroll->scrollwin, A_NORMAL);
}

void get_stats(unsigned long int total_lines, unsigned long int line)
{
time_t now;
struct tm tmptr;
char sdate[64];
static long int page_num = 1;

    if (!(pstat = (struct my_windows *)malloc(sizeof(struct my_windows)))) {
        perror("malloc");
        exit(EXIT_FAILURE);
    }
    pstat->statwin = newwin(1, COLS, LINES - 1, 0);

    time(&now);
    tmptr = *localtime(&now);
    strftime(sdate, sizeof sdate, "%a %b %d  %I:%M:%S%p", &tmptr);

    wbkgd(pstat->statwin, A_REVERSE);
    if (!PIPED || ALONE) {
        if ((line % 23) == 0)
            page_num++;
        mvwprintw(pstat->statwin, 0, 0, "%ld/%ld - Page: %ld", line, total_lines, page_num);
    }
    else 
        mvwprintw(pstat->statwin, 0, 0, "line: %ld", line); 
    mvwprintw(pstat->statwin, 0, COLS - 23, "%s", sdate);

    wrefresh(pstat->statwin);
}

void start_editor(unsigned long int line)
{
char *editor = NULL, *command = NULL;

    if (getenv("VISUAL"))
        editor = getenv("VISUAL");
    else if (getenv("EDITOR")) 
        editor = getenv("EDITOR");
    else
        editor = "/usr/bin/vi";

    command = (char *)malloc(sizeof(char) * strlen(editor) + strlen(filename) + 10);
    sprintf(command, "%s +%ld \"%s\"", editor, line, filename);
    def_prog_mode();
    system(command);
    free(command);
    clear();
    refresh();
    reset_prog_mode();
    refresh();
}

void user_input(unsigned int *scroll_speed, unsigned int origspeed, unsigned
int *toggle, unsigned long int total_lines, unsigned long int line)
{
    switch (get_key()) {
        case 1:
            if (*scroll_speed != 1) /* speed up */
                *scroll_speed = 1;
            else 
                *scroll_speed = origspeed; /* stop speeding */
                break;
        case 2:
            quit_cleanly();
            break;
        case 3:
            *scroll_speed = origspeed;
            break;
        case 4:
            if (!NO_STATUSBAR)
                if (toggle)
                    get_stats(total_lines, line);
            nodelay(stdscr, FALSE);	
            getch();
            break;
        case 5:
            if (NO_STATUSBAR) 
                NO_STATUSBAR = OFF; // 0;
            *toggle = ON;
            break;
        case 6:
            if (NO_STATUSBAR)
                break;
            *toggle = OFF;
            wbkgd(pstat->statwin, A_NORMAL);
            wclear(pstat->statwin);
            wrefresh(pstat->statwin);
            break;
        case 7:
            wclear(pscroll->scrollwin);
            wrefresh(pscroll->scrollwin);
            break;
        case 8: 
            *scroll_speed -= ((*scroll_speed * 25) / 100);
            break;
        case 9:
            *scroll_speed += ((*scroll_speed * 25) / 100);
            break;
        case 10:
            AUTO_PAUSE ^= 1; /* toggle */
            break;
        case 11:
            if (BEEP_OK) 
                beep();
            show_info(*scroll_speed, origspeed, total_lines, line);
            break;
        case 12:
            start_editor(line);
            break;
    }
}

unsigned int get_key(void)
{
chtype key;

    nodelay(stdscr, TRUE);

    if ((key = getch()) != (unsigned long int)ERR)
        if (key == 32)	
            return 1;
        if (key == 'q')
            return 2;
        if (key == 'o')
            return 3;
        if (key == 'p')
            return 4;
        if (key == 'v')
            return 5;
        if (key == 'n')
            return 6;
        if (key == 'c')
            return 7;
        if (key == 'f')
            return 8;
        if (key == 's')
            return 9;
        if (key == 'a')
            return 10;
        if (key == 'i')
            return 11;
        if (key == 'e')
            return 12;

    return 0;
}

void show_info(unsigned int scroll_speed, unsigned int origspeed, unsigned long int total_lines, unsigned long int line)
{
char current_dir[256];
char *header_msg = "Lazyread 1.6  [Press any key]";

    if (!getcwd(current_dir, sizeof current_dir)) {
        refresh(); 
        endwin();
        perror("getcwd()");
        exit(EXIT_FAILURE);
    }
    def_prog_mode(); /* save the old screen */
    clear(); 
    refresh();
 
    box(stdscr, 0, 0);
    mvprintw(0, ((COLS / 2) - (strlen(header_msg) / 2)), "%s", header_msg);
    mvhline(1, 1, ACS_HLINE, COLS - 2);
    mvprintw(2, 1, "Filename: %s", filename);
    mvprintw(3, 1, "File size: %ld bytes", THE_FILE_SIZE);
    mvprintw(4, 1, "Current Dir: %s", current_dir);
    mvprintw(5, 1, "Current line: %ld. Total lines: %ld", line, total_lines);
    mvprintw(6, 1, "Scroll Speed: (%d milliseconds) - %.2f seconds", scroll_speed, (float)scroll_speed / 1000);
    mvprintw(7, 1, "Original Speed: (%d milliseconds) - %.2f seconds", origspeed, (float)origspeed / 1000);
    mvprintw(8, 1, "Terminal Size: %d Rows, %d Columns", LINES, COLS);
    mvprintw(9, 1, "Beep on events: %s", BEEP_OK ? "Yes" : "No");
    mvprintw(10, 1, "Auto-Pause: %s", AUTO_PAUSE ? "Yes" : "No");
    mvprintw(11, 1, "Skip extra blank lines: %s", VIEW_NORMAL ? "No" : "Yes");

    nodelay(stdscr, FALSE);
    getch();
    clear(); 
    refresh();
    reset_prog_mode(); /* Return the screen */
    refresh(); /* needed to show the screen */
    WAS_IN_SHOW_INFO = 1; 
}

void usage()
{
    refresh(); 
    endwin();

    printf("Usage: %s [OPTIONS]\n", progname);
    printf("-f,file <filename>    File to scroll.\n"
    "-s,speed <n>          Scroll at <n> milliseconds. Default: 1000 (1 sec)\n"
    "-pos,position <n>     Start text on row <n> (1 to LINES-2).\n"
    "-w,word <string>      Highlight all lines that <string> appears on.\n"
    "-case <lower|upper>   Display in all lower or all upper case.\n"
    "-color <color>        red, green, blue, bgblue, yellow, magenta.\n"
    "-n,normal             Don't take out extra blank lines. Display as is.\n" 
    "-a,autopause          Automatically pause on highlight word from -w.\n"
    "-v,version            Display program version number.\n"
    "-u,usage              This message.\n"
    "-no-stat-bar          Don't show status bar.\n"
    "-b                    Allow beeping on important events.\n"

    "\tWhile lazyread is running you can use the option keys:\n"
    "'q' to quit\n'p' to pause\n'spacebar' to scroll super"
    "fast (hit again to turn off)\n'c' to clear the current screen"
    "\n'n' to turn off the status bar\n'v' to turn the status bar back on\n"
    "'f' to speed scrolling up in 25 percent increments\n's' to slow "
    "scrolling speed down in 25 percent increments\n"
    "'o' to go back to original speed\n'a' to toggle Auto-Pausing on/off\n"
    "'e' open file for editing in an external editor\n"
    "'i' to view detailed file/program/etc information\n");

    exit(EXIT_FAILURE);
}

void quit_cleanly(void)
{
    refresh();
    flushinp();
    endwin();
    exit(EXIT_SUCCESS);
}

int char_check(char *str)
{
/* checks if any non-digits are found in the input string */

    for (; *str; str++)
        if (!isdigit(*str))
            return 0;

    return 1;
}

long line_nums(FILE *fp)
{
char buf[255];
long linenum = 0;

    /* add up all the lines of the file */
    while (fgets(buf, sizeof(buf), fp))
        ++linenum;

    rewind(fp);

    return (linenum);
}
